import org.apache.commons.lang3.StringUtils

fieldHelper = new FieldHelperTool(apiClientFactory, componentAccessor, i18n, workRatio, xlsContent, baseurl, user)

/**
 * Tool to convert Jira system- and custom fields to their best Excel representations.
 */
public class FieldHelperTool {
	/* Configuration. */
	def issueKeyAsLink = true
	def summaryAsLink = true

	/* Jira Service Management configuration (see https://www.midori-global.com/products/better-excel-exporter-for-jira/cloud/documentation/integrations/jira-service-desk). */
	def exportServiceDeskApprovalDecisions = false // set to "true" to export approver decisions

	/* Insight configuration (see https://www.midori-global.com/products/better-excel-exporter-for-jira/cloud/documentation/integrations/insight). */
	/* Insight in Jira Service Management authentication. */
	def insightUserEmail = '' // Jira Cloud user email address
	def insightApiToken = '' // Jira Cloud API token
	/* Insight Asset Management (legacy app) authentication. */
	def insightAccessToken = '' // Insight Cloud REST API access token

	def exportInsightObjectNames = true // set to "true" to export Insight object names
	def exportInsightObjectKeys = true // set to "true" to export Insight object keys
	def exportInsightObjectAttributes = false // set to "true" to export Insight object attributes

	/* Dependencies. */
	def commentManager = componentAccessor.commentManager
	def customFieldManager = componentAccessor.customFieldManager

	/* Insight dependencies. */
	def insightClient
	def insightInJiraServiceManagementClient

	def apiClientFactory
	def componentAccessor
	def i18n
	def workRatio
	def xlsContent

	def baseUrl
	def user
	def calendar

	private static EXCEL_PARENT_KEY_FIELD_NAME = "Excel parent key"
	private static EXCEL_COMMENTS_CALCULATED_FIELD_NAME = "Excel comments"
	private static EXCEL_FIRST_COMMENT_CALCULATED_FIELD_NAME = "Excel first comment"
	private static EXCEL_FIRST_COMMENT_TEXT_CALCULATED_FIELD_NAME = "Excel first comment text"
	private static EXCEL_FIRST_COMMENT_AUTHOR_CALCULATED_FIELD_NAME = "Excel first comment user"
	private static EXCEL_FIRST_COMMENT_DATE_CALCULATED_FIELD_NAME = "Excel first comment date"
	private static EXCEL_LAST_COMMENT_CALCULATED_FIELD_NAME = "Excel last comment"
	private static EXCEL_LAST_COMMENT_TEXT_CALCULATED_FIELD_NAME = "Excel last comment text"
	private static EXCEL_LAST_COMMENT_AUTHOR_CALCULATED_FIELD_NAME = "Excel last comment user"
	private static EXCEL_LAST_COMMENT_DATE_CALCULATED_FIELD_NAME = "Excel last comment date"

	private static CALCULATED_FIELD_NAMES = [
			EXCEL_PARENT_KEY_FIELD_NAME,
			EXCEL_COMMENTS_CALCULATED_FIELD_NAME,
			EXCEL_FIRST_COMMENT_CALCULATED_FIELD_NAME, EXCEL_FIRST_COMMENT_TEXT_CALCULATED_FIELD_NAME, EXCEL_FIRST_COMMENT_AUTHOR_CALCULATED_FIELD_NAME, EXCEL_FIRST_COMMENT_DATE_CALCULATED_FIELD_NAME,
			EXCEL_LAST_COMMENT_CALCULATED_FIELD_NAME, EXCEL_LAST_COMMENT_TEXT_CALCULATED_FIELD_NAME, EXCEL_LAST_COMMENT_AUTHOR_CALCULATED_FIELD_NAME, EXCEL_LAST_COMMENT_DATE_CALCULATED_FIELD_NAME
	]

	/**
	 * Cell values using structured representation are indented with this.
	 */
	private static INDENTATION = "    "

	FieldHelperTool(def apiClientFactory, def componentAccessor, def i18n, def workRatio, def xlsContent, def baseUrl, def user) {
		this.apiClientFactory = apiClientFactory
		this.componentAccessor = componentAccessor
		this.i18n = i18n
		this.workRatio = workRatio
		this.xlsContent = xlsContent

		this.baseUrl = baseUrl
		this.user = user

		// create calendar with the user's time zone
		def timeZone = this.user.timeZone ?: TimeZone.getDefault()
		calendar = new GregorianCalendar(timeZone)
	}

	/**
	 * Returns whether the passed field ID is related to a custom field (true) or a system field (false).
	 */
	private static boolean isCustomField(String fieldId) {
		fieldId.startsWith("customfield_")
	}

	private static boolean isCustomField(field) {
		isCustomField(field.id)
	}

	/**
	 * Returns whether the passed field name is related to a calculated field.
	 */
	private static boolean isCalculatedField(String customFieldName) {
		customFieldName in CALCULATED_FIELD_NAMES
	}

	private static boolean isCalculatedField(customField) {
		isCalculatedField(customField.name)
	}

	/**
	 * Returns the system-, custom- or calculated field value for the issue.
	 * Note: the returned value is strongly typed (String, Date, etc.).
	 *
	 * @param fieldId is either the ID of a system field (e.g. "assignee"),
	 * 				ID of a custom field (e.g. "customfield_10123"),
	 * 				or the ID of a calculated field (e.g. "Excel parent key").
	 */
	def getFieldValue(issue, String fieldId) {
		if (isCalculatedField(fieldId)) {
			return getCalculatedFieldValue(issue, fieldId)
		}
		if (!isCustomField(fieldId)) {
			return getSystemFieldValue(issue, fieldId)
		}

		def customField = customFieldManager.getCustomFieldObject(fieldId)
		if (customField == null) {
			return "${fieldId} not found"
		}

		if (isCalculatedField(customField)) {
			return getCalculatedFieldValue(issue, customField)
		}

		return getCustomFieldValue(issue, customField)
	}

	def getFieldValue(issue, field) {
		getFieldValue(issue, field.id)
	}

	/**
	 * This convenience method is only to be used if you want to get a field value by the field name.
	 * The #getFieldValue(issue, String fieldId) method is generally a safer option, because that accepts the immutable field ID (e.g. "customfield_10123"), not the name.
	 * This is more convenient because it can work out of the box without configuring the custom field IDs in the templates, but breaks when the custom field is renamed.
	 */
	def getFieldValueByName(issue, String fieldName) {
		def customField = customFieldManager.getCustomFieldObjectByName(fieldName)
		if (customField == null) {
			return "\"${fieldName}\" not found"
		}

		return getCustomFieldValue(issue, customField)
	}

	/**
	 * This convenience method is similar to the previous one, but tolerates custom field renaming by receiving multiple name variants.
	 *
	 * @param fieldNameVariants is an array with multiple custom field names.
	 * 				The value of the first existing custom field name will be returned.
	 */
	def getFieldValueByName(issue, String[] fieldNameVariants) {
		def customField = null

		for (fieldName in fieldNameVariants) {
			customField = customFieldManager.getCustomFieldObjectByName(fieldName)
			if (customField) {
				break
			}
		}
		if (customField == null) {
			return "${fieldNameVariants} not found"
		}

		return getCustomFieldValue(issue, customField)
	}

	private def getSystemFieldValue(issue, fieldId) {
		switch (fieldId) {
			case "project":
				issue.projectObject.name
				break

			case "versions":
				issue.affectedVersions.collect { it.name }.join(", ")
				break

			case "assignee":
				nullSafeUser(issue.assignee)
				break

			case "components":
				issue.components.collect { it.name }.join(", ")
				break

			case "description":
				issue.descriptionRaw
				break

			case "duedate":
				nullSafeDate(issue.dueDate)
				break

			case "environment":
				issue.environmentRaw
				break

			case "fixVersions":
				issue.fixVersions.collect { it.name }.join(", ")
				break

			case "issuekey":
				issueToLink(issue)
				break

			case "parent":
				issue.parentObject ? issueToLink(issue.parentObject) : null
				break

			case "issuetype":
				issue.issueTypeObject.nameTranslation
				break

			case "issuelinks":
				def linkCollection = xlsContent.linkCollectionByIssue(issue)
				linkCollection.allIssues.collect { it.key }.join(", ")
				break

			case "lastViewed":
				nullSafeDateTime(issue.lastViewed)
				break

			case "workratio":
				def percentage = workRatio.getWorkRatio(issue)
				(percentage >= 0) ? (percentage / 100.0) : ""
				break

			case "subtasks":
				issue.subTaskObjects.collect { it.key }.join(", ")
				break

			case "priority":
				issue.priorityObject.nameTranslation
				break

			case "reporter":
				nullSafeUser(issue.reporter)
				break

			case "creator":
				nullSafeUser(issue.creator)
				break

			case "security":
				issue.securityLevel?.name
				break

			case "summary":
				summaryAsLink ? "${issue.summary}|${baseUrl}/browse/${issue.key}" : issue.summary
				break

			case "created":
				nullSafeDateTime(issue.getCreated())
				break

			case "updated":
				nullSafeDateTime(issue.updated)
				break

			case "resolutiondate":
				nullSafeDateTime(issue.resolutionDate)
				break

			case "status":
				issue.statusObject.nameTranslation
				break

			case "statusCategory":
				issue.statusObject.statusCategory.nameTranslation
				break

			case "statuscategorychangedate":
				nullSafeDateTime(issue.statusCategoryChangeDate)
				break

			case "resolution":
				issue.resolutionObject?.name
				break

			case "labels":
				issue.labels.collect { it.label }.join(", ")
				break

			case "timeoriginalestimate":
				secondsToHours(issue.originalEstimate)
				break

			case "timeestimate":
				secondsToHours(issue.estimate)
				break

			case "timespent":
				secondsToHours(issue.timeSpent)
				break

			case "aggregatetimespent":
				secondsToHours(issue.aggregateTimeSpent)
				break

			case "aggregatetimeestimate":
				secondsToHours(issue.aggregateRemainingEstimate)
				break

			case "aggregatetimeoriginalestimate":
				secondsToHours(issue.aggregateOriginalEstimate)
				break

			case "aggregateprogress":
				def percentage = xlsContent.calculateProgressPercentage(issue.aggregateTimeSpent, issue.aggregateRemainingEstimate)
				(percentage >= 0) ? (percentage / 100.0) : ""
				break

			case "progress":
				def percentage = xlsContent.calculateProgressPercentage(issue.timeSpent, issue.estimate)
				(percentage >= 0) ? (percentage / 100.0) : ""
				break

			case "votes":
				issue.votes
				break

			case "watches":
				issue.watches
				break

			default:
				// other types, including 'comment', 'thumbnail', 'attachment', 'timetracking', 'worklog', 'issuerestriction'
				"${fieldId} not supported yet"
				break
		}
	}

	private def getCustomFieldValue(issue, customField) {
		def customFieldTypeKey = customField.customFieldType.key
		def value = issue.getCustomFieldValue(customField)
		def valueRaw = issue.getCustomFieldValueRaw(customField)

		switch (customFieldTypeKey) {
			case "com.atlassian.jira.plugin.system.customfieldtypes:multicheckboxes":
			case "com.atlassian.jira.plugin.system.customfieldtypes:multiselect":
				(value != null) ? value.collect { it.value }.join(", ") : null
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:datepicker":
				nullSafeDate(value)
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:datetime":
				nullSafeDateTime(value)
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:multigrouppicker":
			case "com.atlassian.jira.plugin.system.customfieldtypes:grouppicker":
			case "com.atlassian.jira.plugin.system.customfieldtypes:multiversion":
			case "com.atlassian.jira.plugin.system.customfieldtypes:version":
				(value != null) ? value.collect { it.name }.join(", ") : null
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:labels":
				(value != null) ? value.collect { it.label }.join(", ") : null
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:float":
				value // to keep its original Double type (not to convert it to String)
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:project":
				value?.name
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:radiobuttons":
			case "com.atlassian.jira.plugin.system.customfieldtypes:select":
				value?.value
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:cascadingselect":
				(value != null) ? value.get(null).toString() + ((value.get("1") != null) ? " - " + value.get("1").toString() : "") : null
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:readonlyfield":
			case "com.atlassian.jira.plugin.system.customfieldtypes:textarea":
			case "com.atlassian.jira.plugin.system.customfieldtypes:textfield":
			case "com.atlassian.jira.plugin.system.customfieldtypes:url":
				valueRaw
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:multiuserpicker":
				(value != null) ? value.collect { nullSafeUser(it) }.join(", ") : null
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:userpicker":
				nullSafeUser(value)
				break

			// Jira Software custom field types
			case "com.pyxis.greenhopper.jira:gh-sprint":
				(value != null) ? value.collect { it.name }.join(", ") : null
				break

			case "com.pyxis.greenhopper.jira:gh-epic-color":
			case "com.pyxis.greenhopper.jira:gh-epic-label":
			case "com.pyxis.greenhopper.jira:gh-epic-status":
			case "com.pyxis.greenhopper.jira:jsw-issue-color":
			case "com.pyxis.greenhopper.jira:jsw-story-points":
				value
				break

			case "com.pyxis.greenhopper.jira:gh-epic-link":
				value ? customField.customFieldType.getEpicDisplayName(value) : null
				break

			// Jira Charting Plugin custom field types
			case "com.atlassian.jira.ext.charting:firstresponsedate":
			case "com.atlassian.jira.ext.charting:timeinstatus":
				"${customFieldTypeKey} not supported yet"
				break

			// Jira Development custom field types
			case "com.atlassian.jira.plugins.jira-development-integration-plugin:devsummarycf":
				"${customFieldTypeKey} not supported yet"
				break

			// Jira Toolkit custom field types
			case "com.atlassian.jira.toolkit:assigneedomain":
			case "com.atlassian.jira.toolkit:lastusercommented":
			case "com.atlassian.jira.toolkit:message":
			case "com.atlassian.jira.toolkit:reporterdomain":
			case "com.atlassian.jira.toolkit:userproperty":
			case "com.atlassian.jira.toolkit:viewmessage":
				value
				break

			case "com.atlassian.jira.toolkit:attachments":
			case "com.atlassian.jira.toolkit:comments":
				nullSafeNumber(value)
				break

			case "com.atlassian.jira.toolkit:dayslastcommented":
				nullSafeDate(value)
				break

			case "com.atlassian.jira.toolkit:LastCommentDate":
				nullSafeDateTime(value)
				break

			case "com.atlassian.jira.toolkit:lastupdaterorcommenter":
				nullSafeUser(value)
				break

			case "com.atlassian.jira.toolkit:participants":
				(value != null) ? value.collect { nullSafeUser(it) }.join(", ") : null
				break

			// Advanced Roadmaps custom field types
			case "com.atlassian.jpo:jpo-custom-field-baseline-start":
			case "com.atlassian.jpo:jpo-custom-field-baseline-end":
				nullSafeDate(value)
				break

			case "com.atlassian.jpo:jpo-custom-field-parent":
				value?.data ? issueToLink(value.data, true) : null
				break

			case "com.atlassian.teams:rm-teams-custom-field-team":
				value?.title
				break

			// Jira Service Management custom field types
			case "com.atlassian.servicedesk.approvals-plugin:sd-approvals":
				approvalsToString(value)
				break

			case "com.atlassian.servicedesk.assets-plugin:assetfield": // Insight Asset Management (legacy app) custom field type
				assetsToString(value, customFieldTypeKey)
				break

			case "com.atlassian.servicedesk.servicedesk-lingo-integration-plugin:sd-request-language":
				value?.displayName
				break

			case "com.atlassian.servicedesk:sd-customer-organizations":
				value?.collect { it.name }?.join(", ")
				break

			case "com.atlassian.servicedesk:sd-request-feedback":
				value?.rating
				break

			case "com.atlassian.servicedesk:sd-request-feedback-date":
				nullSafeDateTime(value)
				break

			case "com.atlassian.servicedesk:sd-request-participants":
				value?.collect { nullSafeUser(it) }?.join(", ")
				break

			case "com.atlassian.servicedesk:sd-sla-field":
				slaToHours(value)
				break

			case "com.atlassian.servicedesk:vp-origin":
				value?.requestType?.name
				break

			case "com.atlassian.jira.plugins.cmdb:cmdb-object-cftype": // Insight in Jira Service Management custom field type
				insightObjectsToString(value, customFieldTypeKey)
				break

			case "com.atlassian.jira.plugins.service-entity:service-entity-field-cftype": // Insight in Jira Service Management custom field type ("Affected Services")
				"${customFieldTypeKey} not supported yet"
				break

			// Tempo custom field types
			case "com.atlassian.plugins.atlassian-connect-plugin:io.tempo.jira__account":
			case "com.atlassian.plugins.atlassian-connect-plugin:io.tempo.jira__team":
				value?.value
				break

			default:
				value
				break
		}
	}

	private def getCalculatedFieldValue(issue, customField) {
		return getCalculatedFieldValue(issue, customField.name)
	}

	private def getCalculatedFieldValue(issue, String fieldName) {
		def comments = issue.commentsRaw

		switch (fieldName) {
			case EXCEL_PARENT_KEY_FIELD_NAME:
				issue.isSubTask() ? issueToLink(issue.parentObject) : ""
				break

			case EXCEL_COMMENTS_CALCULATED_FIELD_NAME:
				!comments.empty ? comments.collect { commentToString(it) }.join("\n\n") : ""
				break

			case EXCEL_FIRST_COMMENT_CALCULATED_FIELD_NAME:
				!comments.empty ? commentToString(comments.first()) : ""
				break

			case EXCEL_FIRST_COMMENT_TEXT_CALCULATED_FIELD_NAME:
				!comments.empty ? comments.first().body : ""
				break

			case EXCEL_FIRST_COMMENT_AUTHOR_CALCULATED_FIELD_NAME:
				!comments.empty ? comments.first().authorFullName : ""
				break

			case EXCEL_FIRST_COMMENT_DATE_CALCULATED_FIELD_NAME:
				!comments.empty ? nullSafeDateTime(comments.first().created) : ""
				break

			case EXCEL_LAST_COMMENT_CALCULATED_FIELD_NAME:
				!comments.empty ? commentToString(comments.last()) : ""
				break

			case EXCEL_LAST_COMMENT_TEXT_CALCULATED_FIELD_NAME:
				!comments.empty ? comments.last().body : ""
				break

			case EXCEL_LAST_COMMENT_AUTHOR_CALCULATED_FIELD_NAME:
				!comments.empty ? comments.last().authorFullName : ""
				break

			case EXCEL_LAST_COMMENT_DATE_CALCULATED_FIELD_NAME:
				!comments.empty ? nullSafeDateTime(comments.last().created) : ""
				break

			default:
				"${fieldName} is an undefined calculated field"
				break
		}
	}

	/**
	 * Returns the Excel cell type for the issue field: "number", "date" or "string".
	 */
	def getFieldCellType(String fieldId) {
		if (isCalculatedField(fieldId)) {
			return getCalculatedFieldCellType(fieldId)
		}
		if (!isCustomField(fieldId)) {
			return getSystemFieldCellType(fieldId)
		}

		def customField = customFieldManager.getCustomFieldObject(fieldId)
		if (customField == null) {
			return "string"
		}

		if (isCalculatedField(customField)) {
			return getCalculatedFieldCellType(customField)
		}

		return getCustomFieldCellType(customField)
	}

	def getFieldCellType(field) {
		getFieldCellType(field.id)
	}

	private def getSystemFieldCellType(fieldId) {
		switch (fieldId) {
			case "workratio":
			case "timeoriginalestimate":
			case "timeestimate":
			case "timespent":
			case "aggregatetimespent":
			case "aggregatetimeestimate":
			case "aggregatetimeoriginalestimate":
			case "aggregateprogress":
			case "progress":
			case "votes":
			case "watches":
				"number"
				break

			case "duedate":
			case "created":
			case "updated":
			case "resolutiondate":
			case "statuscategorychangedate":
			case "lastViewed":
				"date"
				break

			case "issuekey":
			case "parent":
				issueKeyAsLink ? "link" : "string"
				break

			case "summary":
				summaryAsLink ? "link" : "string"
				break

			default:
				"string"
		}
	}

	private def getCustomFieldCellType(customField) {
		switch (customField.customFieldType.key) {
			case "com.atlassian.jira.plugin.system.customfieldtypes:datepicker":
			case "com.atlassian.jira.plugin.system.customfieldtypes:datetime":
			case "com.atlassian.jira.toolkit:dayslastcommented":
			case "com.atlassian.jira.toolkit:LastCommentDate":
			case "com.atlassian.jpo:jpo-custom-field-baseline-start":
			case "com.atlassian.jpo:jpo-custom-field-baseline-end":
			case "com.atlassian.servicedesk:sd-request-feedback-date":
				"date"
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:float":
			case "com.atlassian.jira.toolkit:attachments":
			case "com.atlassian.jira.toolkit:comments":
			case "com.atlassian.servicedesk:sd-request-feedback":
			case "com.atlassian.servicedesk:sd-sla-field":
			case "com.pyxis.greenhopper.jira:jsw-story-points":
				"number"
				break

			case "com.atlassian.jpo:jpo-custom-field-parent":
				issueKeyAsLink ? "link" : "string"
				break

			default:
				"string"
		}
	}

	private def getCalculatedFieldCellType(customField) {
		return getCalculatedFieldCellType(customField.name)
	}

	private def getCalculatedFieldCellType(String fieldName) {
		switch (fieldName) {
			case EXCEL_PARENT_KEY_FIELD_NAME:
				issueKeyAsLink ? "link" : "string"
				break

			case EXCEL_FIRST_COMMENT_DATE_CALCULATED_FIELD_NAME:
			case EXCEL_LAST_COMMENT_DATE_CALCULATED_FIELD_NAME:
				"date"
				break

			default:
				"string"
		}
	}

	/**
	 * Returns the format specifier string (ex: "dd/mmm/yyyy") for the issue field.
	 * The format specifier is typically used for number- or date type fields.
	 */
	def getFieldFormat(String fieldId) {
		if (isCalculatedField(fieldId)) {
			return getCalculatedFieldFormat(fieldId)
		}
		if (!isCustomField(fieldId)) {
			return getSystemFieldFormat(fieldId)
		}

		def customField = customFieldManager.getCustomFieldObject(fieldId)
		if (customField == null) {
			return null
		}

		if (isCalculatedField(customField)) {
			return getCalculatedFieldFormat(customField)
		}

		return getCustomFieldFormat(customField)
	}

	def getFieldFormat(field) {
		getFieldFormat(field.id)
	}

	private def getSystemFieldFormat(fieldId) {
		switch (fieldId) {
			case "workratio":
			case "progress":
			case "aggregateprogress":
				"0.00%"
				break

			case "timeoriginalestimate":
			case "timeestimate":
			case "timespent":
			case "aggregatetimespent":
			case "aggregatetimeestimate":
			case "aggregatetimeoriginalestimate":
				"0.00"
				break

			case "duedate":
			case "created":
			case "updated":
			case "resolutiondate":
			case "statuscategorychangedate":
				"dd/mmm/yyyy"
				break

			case "lastViewed":
				"dd/mmm/yyyy hh:mm"
				break

			default:
				// no custom format
				null
		}
	}

	private def getCustomFieldFormat(customField) {
		switch (customField.customFieldType.key) {
			case "com.atlassian.jira.plugin.system.customfieldtypes:datepicker":
			case "com.atlassian.jira.toolkit:dayslastcommented":
			case "com.atlassian.jpo:jpo-custom-field-baseline-start":
			case "com.atlassian.jpo:jpo-custom-field-baseline-end":
				"dd/mmm/yyyy"
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:datetime":
			case "com.atlassian.jira.toolkit:LastCommentDate":
			case "com.atlassian.servicedesk:sd-request-feedback-date":
				"dd/mmm/yyyy hh:mm"
				break

			case "com.atlassian.servicedesk:sd-sla-field":
				"0.00\" h\";[RED]-0.00\" h\""
				break

			default:
				// no custom format
				null
		}
	}

	private def getCalculatedFieldFormat(customField) {
		return getCalculatedFieldFormat(customField.name);
	}

	private def getCalculatedFieldFormat(String fieldName) {
		switch (fieldName) {
			case EXCEL_FIRST_COMMENT_DATE_CALCULATED_FIELD_NAME:
			case EXCEL_LAST_COMMENT_DATE_CALCULATED_FIELD_NAME:
				"dd/mmm/yyyy hh:mm"
				break

			default:
				// no custom format
				null
		}
	}

	/**
	 * Returns the format specifier string (ex: "[RED]") for the issue field.
	 * Use this method only in those rare cases when the format specifier string depends on the concrete field value,
	 * not only on the field itself. If it doesn't, use {@link #getFieldFormat(java.lang.Object)} for better performance.
	 */
	def getFieldDynamicFormat(def issue, String fieldId) {
		if (isCalculatedField(fieldId)) {
			return getCalculatedFieldDynamicFormat(issue, fieldId)
		}
		if (!isCustomField(fieldId)) {
			return getSystemFieldDynamicFormat(issue, fieldId)
		}

		def customField = customFieldManager.getCustomFieldObject(fieldId)
		if (customField == null) {
			return null
		}

		if (isCalculatedField(customField)) {
			return getCalculatedFieldDynamicFormat(issue, customField.name)
		}

		return getCustomFieldDynamicFormat(issue, customField)
	}

	private def getSystemFieldDynamicFormat(issue, fieldId) {
		switch (fieldId) {
			case "status":
			case "statusCategory":
				// export status and status category in color (see http://dmcritchie.mvps.org/excel/colors.htm)
				def statusCategoryColor = issue.status?.statusCategory?.colorName
				switch (statusCategoryColor) {
					case "blue-gray":
						return "[BLACK]" // black
					case "green":
						return "[Color10]" // dark green
					case "yellow":
						return "[Color46]" // dark orange
				}
				break

			case "duedate":
				// export expired due date for unresolved issues in red (see http://dmcritchie.mvps.org/excel/colors.htm)
				def isDueDateExpired = (issue.dueDate != null) && (issue.dueDate < new Date())
				def isResolved = (issue.status.statusCategory.key == "done") || (issue.resolution != null)
				def color = (isDueDateExpired && !isResolved) ? "[RED]" : ""
				"${color}dd/mmm/yyyy"
				break

			default:
				// no custom format
				null
		}
	}

	private def getCalculatedFieldDynamicFormat(issue, fieldName) {
		switch (fieldName) {
			default:
				// no custom format
				null
		}
	}

	private def getCustomFieldDynamicFormat(issue, customField) {
		switch (customField.customFieldType.key) {
			default:
				// no custom format
				null
		}
	}

	/**
	 * Returns the issue key (optionally with the summary) as plain text or as clickable link.
	 */
	private def issueToLink(issue, showSummary = false) {
		def text = issue.key + (showSummary ? " " + issue.summary : "")
		return issueKeyAsLink ? (text + "|${baseUrl}/browse/" + issue.key) : text
	}

	/**
	 * Returns the string that contains the comment text, the comment author and the comment date.
	 */
	private def commentToString(def comment) {
		def commentString = comment.body + "\n---"

		def commentUser = nullSafeUser(comment.authorApplicationUser)
		if (!"".equals(commentUser)) {
			commentString += " by " + commentUser
		}

		def commentCalendar = nullSafeDateTime(comment.created)
		if (!"".equals(commentCalendar)) {
			commentString += " at " + commentCalendar.format("dd/MMM/yyyy HH:mm")
		}

		return commentString
	}

	/**
	 * Should be used for the values without time component and time zone.
	 */
	private def nullSafeDate(def date) {
		(date != null) ? date : ""
	}

	/**
	 * Should be used for the values with time component and time zone.
	 */
	def nullSafeDateTime(def date) {
		if (date != null) {
			calendar.setTime(date)
			return calendar
		} else {
			return ""
		}
	}

	private static def nullSafeUser(def user) {
		(user != null) ? user.displayName : ""
	}

	/**
	 * Returns the value or 0 if the value is <code>null</code>.
	 */
	private static def nullSafeNumber(def number) {
		(number != null) ? number : 0
	}

	private static Double secondsToHours(def seconds) {
		(seconds != null) ? (seconds / (60.0 * 60.0)) : null
	}

	/**
	 * Returns the body text fragments from the passed XML joined with the delimiter.
	 * The XML is not required to be well-formatted and can even be plain text.
	 */
	private static xmlToDelimitedString(def xml, def delimiter) {
		xml.split(/<.*?>/).collect { it.trim() }.findAll { !it.allWhitespace }.join(delimiter)
	}

	// - Jira Service Management integration ----------------------------------

	private def getApprovalDecision(def approvalDecision) {
		if (approvalDecision == "approved") {
			return i18n.getText("sd.approval.status.approved")
		} else if (approvalDecision == "declined") {
			return i18n.getText("sd.approval.status.declined")
		} else {
			return i18n.getText("sd.approval.status.pending")
		}
	}

	private def approvalsToString(def value) {
		value?.collect { approval ->
			def decisionsAsString = ''
			if (exportServiceDeskApprovalDecisions) {
				def decisions = approval.approvers.collect { approver -> "${approver.approver.displayName}: ${getApprovalDecision(approver.approverDecision)}" }.join(", ")
				decisionsAsString = " (${decisions})"
			}
			"${approval.name}: ${getApprovalDecision(approval.finalDecision)}${decisionsAsString}"
		}?.join(", ")
	}

	private def slaToHours(def value) {
		def cycle = value?.ongoingCycle ?: value?.completedCycles?.last()
		def millis = cycle?.remainingTime?.millis
		millis ? (millis / (60*60*1000)) : null
	}

	// - Insight integration --------------------------------------------------

	private def assetsToString(def value, def customFieldTypeKey) {
		def objectAppKey = value ? value.get(0).appKey : ''
		if (objectAppKey == 'com.riadalabs.jira.plugins.insight') {
			insightObjectsToString(value, customFieldTypeKey)
		} else {
			objectAppKey ? "\"${objectAppKey}\" assets not supported yet" : null
		}
	}

	private def insightObjectsToString(def value, def customFieldTypeKey) {
		try {
			def objects = fetchInsightObjects(value, customFieldTypeKey)
			return objects.sort { it.name }.collect { object ->
				def nameAsString = exportInsightObjectNames ? object.name : ''
				def objectKeysAsString = exportInsightObjectKeys ? " [${object.objectKey}]" : ''
				def attributesAsString = exportInsightObjectAttributes ? " (${insightAttributesToString(object)})" : ''
				"${nameAsString}${objectKeysAsString}${attributesAsString}"
			}.findAll { it != '' }.join(', ')
		} catch (Throwable ex) {
			return "Failed to download Insight object data! (see Insight API guide: https://www.midori-global.com/products/better-excel-exporter-for-jira/cloud/documentation/integrations/insight): ${ex.message}"
		}
	}

	private def fetchInsightObjects(def customFieldValue, def customFieldTypeKey) {
		if (!customFieldValue) {
			return []
		}

		def appType = (customFieldTypeKey != "com.atlassian.servicedesk.assets-plugin:assetfield") ? "jsm" : "legacy"
		if ((appType == "jsm" && (!insightUserEmail || !insightApiToken)) || (appType == "legacy" && !insightAccessToken)) {
			throw new IllegalStateException("Enable Insight!")
		}
		if (appType == "jsm") {
			if (!insightInJiraServiceManagementClient) {
				def workspaceId = customFieldValue[0].workspaceId
				insightInJiraServiceManagementClient = apiClientFactory.createInsightInJiraServiceManagementClient(workspaceId, insightUserEmail, insightApiToken)
			}
			def objectIds = customFieldValue.collect { it.objectId }
			return insightInJiraServiceManagementClient.getInsightObjects(objectIds)
		} else {
			if (!insightClient) {
				insightClient = apiClientFactory.createInsightClient(insightAccessToken)
			}
			def objectIds = customFieldValue.collect { StringUtils.substringAfterLast(it.originId, "_") }
			return insightClient.getInsightObjects(objectIds)
		}
	}

	private static def insightAttributesToString(def object) {
		object.attributes
				.sort { a, b -> a.objectTypeAttribute.position <=> b.objectTypeAttribute.position }
				.collect {
					attribute ->
						def attributeValues = attribute.objectAttributeValues.collect { it.displayValue }.join(';')
						"${attribute.objectTypeAttribute.name}=${attributeValues}"
				}.join(', ')
	}
}
