import org.apache.commons.lang3.StringUtils

filename = "${title} - projects.xlsx".toString() // "filename" is a standard Java String, not a GString

projectExport = new ProjectExportTool(componentAccessor: componentAccessor, apiClientFactory: apiClientFactory)

/**
 * Tool to export Jira projects.
 */
class ProjectExportTool {
	def componentAccessor
	def apiClientFactory

	def jiraClient
	def projectIdsToIssueCounts = [:]
	def projectIdsToUpdatedDates = [:]

	def slurper = new groovy.json.JsonSlurper()

	def getAllProjects() {
		jiraClient = apiClientFactory.createJiraClient()

		def projects = componentAccessor.projectManager.projectObjects.findAll { !it.private }
		initProjectStatistics(projects)
		return projects
	}

	def getStatus(project) {
		project.archived ? 'Archived' : (project.deleted ? 'Deleted' : 'Active')
	}

	def getProjectType(project) {
		projectTypeKeyToProjectType(project.projectTypeKey.key)
	}

	def getProjectCategory(project) {
		project.projectCategory?.name
	}

	def getLastIssueUpdate(project) {
		projectIdsToUpdatedDates[project.id]
	}

	def getIssueCount(project) {
		projectIdsToIssueCounts[project.id]
	}

	def getComponents(project) {
		project.components.collect { it.name }.join(', ')
	}

	def getIssueTypes(project) {
		project.issueTypes.collect { it.name }.join(', ')
	}

	def getVersions(project) {
		project.versions.collect { it.name }.join(', ')
	}

	/**
	 * Collects statistics for the non-deleted and non-archived projects.
	 */
	private def initProjectStatistics(projects) {
		projects.findAll { !it.deleted && !it.archived }.each {
			try {
				def result = callJiraApi("/search?jql=project=${it.key} ORDER BY updated DESC&maxResults=1")

				projectIdsToUpdatedDates[it.id] = result.issues ? Date.parse("yyyy-MM-dd'T'HH:mm:ss", result.issues[0].fields.updated) : "Never"
				projectIdsToIssueCounts[it.id] = result.total ?: 0
			} catch (Exception ex) {
				projectIdsToUpdatedDates[it.id] = ex.message
			}
		}
	}

	/**
	 * Calls the Jira Cloud REST API.
	 */
	private def callJiraApi(url) {
		def json = jiraClient.getJson(url)
		return slurper.parseText(json)
	}

	/**
	 * Same as in <i>com.atlassian.jira.project.type.ProjectTypeKeyFormatter</i> in Server version.
	 */
	private static def projectTypeKeyToProjectType(projectTypeKey) {
		// handle rebranding
		if (projectTypeKey == 'service_desk')
			projectTypeKey = 'service_management'
		projectTypeKey ? StringUtils.capitalize(StringUtils.replace(projectTypeKey, '_', ' ').trim()) : 'Unknown'
	}
}